/*
  mergemem.c
  
  This file is part of mergemem by Philipp Richter & Philipp Reisner

  mergemem is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  mergemem is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with mergemem; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include "mergemem.h"

static struct CmdList *cmd_list = NULL;
static struct PidList *pid_list = NULL;
static struct DontMerge *dont_merge = NULL;
static char logfile[PATH_MAX] = "/var/log/mergemem.log";
static int loglevel = 3;
static unsigned long int interval = 60;
static int run_as_daemon = 0;
static FILE *logf = NULL;

static int saved_pages;
static int chk_runs;
static int sum_runs;

static int mergemod_fd;
/* --__--__--__--__--__--__--__-- */

int main( int argc, char **argv )
{
  /* initialize parameters */
  if( argc == 1)
  {
    run_as_daemon = 1;
    read_cfg_file();
  }
  else
    parse_options( argc, argv );
  /* open logfile*/
  if( run_as_daemon && (logf = fopen(logfile, "a")) == 0 )
  {
    setvbuf(logf, 0, _IONBF, 0);
    perror("could not open logfile for writing\n");
    free_all();
    exit(1);
  }

  /* check if we are already running */
  if( run_as_daemon) 
    check_running();

  /* open mergemod file */
  if( (mergemod_fd = open("/proc/mergemem", O_WRONLY)) == -1 )
  {
    logmsg( 1, "/proc/mergemem not found. Insert mergemod.o into kernel\n" );
    err_exit("/proc/mergemem not found. Insert mergemod.o into kernel\n");
  }

  logmsg(1, "mergemem started\n");
  if(run_as_daemon)
  {
    pid_t p;
    p = fork();
    if( p == -1 )
      perror("could not fork\n");
    if(p == 0)
    {
      signal(SIGHUP, &cfg_reread);
      signal(SIGTERM, &ok_exit);
      wait_loop();
    }
  }
  else
  {
    mrg_singleshot();
    ok_exit();
  }

  return( 0 );
}

void usage( char *name )
{
  printf( "Usage: %s [options]...\n\n", name );
  printf( "Call kernel module to merge memory of equal commands\n\n"
	  "  -d, --daemon             run in daemon mode (uses config file for finetuning\n"
	  "                           see manpage for details)\n"
	  "  -p PID, --pid PID        use pid PID as one of the pids to merge. at least 2\n"
	  "                           or more times required. the pids must point to\n"
	  "                           different instances of the same command\n"
	  "  -n CMD, --name CMD       use all pids refering to the command CMD\n"
	  "  -l LEV, --loglevel LEV   set log level to LEVEL (1,2,3)\n"
	  "  -f FILE, --logfile FILE  logging goes to file FILE\n"
	  "  -h, --help               print this help and exit\n"
	  "  -v, --version            print version and exit\n"  );
  exit( 0 );
}

void parse_options( int argc, char **argv )
{
  int c;
  char *t;
  struct PidList *pl;
  int p;
  struct CmdList *cl;
  int lev;
  
  static struct option long_options[] =
  {
    { "daemon",   0, 0, 'd' },
    { "pid",      1, 0, 'p' },
    { "name",     1, 0, 'n' },
    { "loglevel", 1, 0, 'l' },
    { "logfile",  1, 0, 'f' },
    { "version",  0, 0, 'v' },
    { "help",     0, 0, 'h' },
    { 0, 0, 0, 0 }
  };

  while(1)
  {
    c = getopt_long( argc, argv, "dp:n:l:f:vh", long_options, 0 );
    if( c == -1 )
      break;

    switch( c )
    {
      case 'd':
	run_as_daemon = 1;
	read_cfg_file();
	break;
      case 'p':
	if((p = atoi(optarg)) < 2)
	  err_exit("pid must be greater that 1\n");
	if((pl = malloc(sizeof(struct PidList))) == 0)
	  err_exit("Memory exhausted\n");                     /* aaaaaarrrggghhh  */
	pl->pid = p;
	pl->next = pid_list;
	pid_list = pl;
	break;
      case 'n':
	if( (cl = malloc(sizeof(struct CmdList) + strlen(optarg) + 1)) == NULL )
	  err_exit("Memory exhausted\n");
	cl->cmd = (char*)cl+sizeof(struct CmdList);
	strcpy( cl->cmd, optarg );
	cl->iv = 0;
	cl->cntr = 0;
	cl->next = cmd_list;
	cmd_list = cl;
	break;
      case 'l':
	if((lev = atoi(optarg)) < 1 || lev > 3)
	  err_exit("loglevel must be between 1 and 3\n");
	loglevel = lev;
	break;
      case 'f':
	strncpy(logfile, optarg, PATH_MAX-1);
	logfile[PATH_MAX] = '\0';
	break;
      case 'v':
	t = strrchr( *argv, '/')+1;
	if( t == (char*)1 ) t = *argv;
	printf( "This is version " VERSION " of %s.\n"
		"Report bugs and errors to\n"
		"e9426459@stud2.tuwien.ac.at or\n"
		"e9525415@stud2.tuwien.ac.at\n", t );
	exit(0);
      default:
	err_exit("Unknown option ...\n");
      case 'h':
      case '?':
	usage( *argv );
    }
  }

  if( optind < argc )
    err_exit( "Too much arguments ...\n" );
}

/*
  read the config file /etc/mergemem.cfg
*/
void read_cfg_file()
{
  FILE *f;
  char buf[PATH_MAX], *buf1, *buf2, *buf3;
  int line;

  /* no cfg file? */
  if( !(f = fopen( CFG_NAME, "r")) )
    return;

  line = 0;
  while( fgets(buf, sizeof(buf), f) ) 
  {
    buf1 = buf;
    line++;
    if( (buf2 = cfg_next_token( &buf1 )) == NULL) continue;
    if( strchr( "#%\n", *buf1) ) continue;

    if( strcmp(buf1, "command") == 0)
    {
      struct CmdList *cl;
      unsigned long int iv;
      if( (buf1 = cfg_next_token( &buf2 )) == NULL)
	cfg_error(line, "argument 1 for 'command' missing");
      if( (buf3 = cfg_next_token( &buf1 )) == NULL)
	cfg_error(line, "argument 2 for 'command' missing");
      if( (iv = atol(buf1)) < 1 )
	cfg_error(line, "interval must be greater than zero");
      /* buf2 = name, buf1 = interval */
      if( (cl = malloc(sizeof(struct CmdList) + strlen(buf2) + 1)) == NULL )
	  err_exit("Memory exhausted\n");
      cl->cmd = (char*)cl+sizeof(struct CmdList);
      strcpy( cl->cmd, buf2 );
      cl->iv = iv;
      cl->cntr = 0;
      cl->next = cmd_list;
      cmd_list = cl;
    } else if( strcmp(buf1, "dontmerge") == 0)
    {
      int foundone = 0;
      struct DontMerge *dm;
      while( (buf1 = cfg_next_token( &buf2 )) )
      {
	if( (dm = malloc(sizeof(struct DontMerge) + strlen(buf2) + 1)) == NULL)
	  err_exit("Memory exhausted\n");
	dm->cmd = (char*)dm+sizeof(struct DontMerge);
	strcpy( dm->cmd, buf2 );
	dm->next = dont_merge;
	dont_merge = dm;
	foundone = 1;
	buf2 = buf1;
      }
      if( !foundone )
	cfg_error(line, "argument(s) for 'dontmerge' missing");
    } else if( strcmp(buf1, "interval") == 0)
    {
      unsigned long int iv;
      if( (buf1 = cfg_next_token( &buf2 )) == NULL)
	cfg_error(line, "argument for 'interval' missing");
      if( (iv = atol(buf2)) < 1 )
	cfg_error(line, "interval must be greater than zero");
      interval = iv;
    } else if( strcmp(buf1, "logfile") == 0)
    {
      if( (buf1 = cfg_next_token( &buf2 )) == NULL)
	cfg_error(line, "argument for 'logfile' missing");
      strncpy( logfile, buf2, PATH_MAX - 1);
      logfile[PATH_MAX] = '\0';
    } else if( strcmp(buf1, "loglevel") == 0)
    {
      int lev;
      if( (buf1 = cfg_next_token( &buf2 )) == NULL)
	cfg_error(line, "argument for 'loglevel' missing");
      if( ( lev = atoi(buf2) )<1 || lev>3 )
	cfg_error(line, "'loglevel' must be between 1 and 3");
      loglevel = lev;
    } else
      cfg_error(line, "Illegal tag");
  }

  fclose(f);
}

/*
  print error message and abort
*/
void cfg_error( int line, const char *text )
{
  /*  printf("Error in %s, line %d: %s\n", CFG_NAME, line, text);*/
  logmsg(1, "Error in %s, line %d: %s\n", CFG_NAME, line, text);
  free_all();
  exit(1);
}

/*
  find next token in str, skip spaces end set a \0 after last nonspace char
*/
char *cfg_next_token( char **str )
{
  char *end;
  while( **str && isspace(**str) ) { (*str)++; }
  end = *str;
  while( *end && !isspace( *end) ) end++;
  if( *str == end ) return NULL;
  *end++ = '\0';
  return end;
}

/*
  free all malloced memory
*/
void free_all()
{
  struct DontMerge *dm;
  struct CmdList *cl;
  struct PidList *pl;
  while(dont_merge)
  {
    dm = dont_merge->next;
    free(dont_merge);
    dont_merge = dm;
  }
  while(cmd_list)
  {
    cl = cmd_list->next;
    free(cmd_list);
    cmd_list = cl;
  }
  while(pid_list)
  {
    pl = pid_list->next;
    free(pid_list);
    pid_list = pl;
  }
}

void err_exit( const char *msg )
{
  if( mergemod_fd != -1 )
    close( mergemod_fd );
  
  printf(msg);
  free_all();
  if(logf)
  {
    logmsg(1, "FATAL ERROR: %s. program will exit now", msg);
    fclose(logf);
  }
  exit(1);
}

void ok_exit()
{
  if( mergemod_fd != -1 )
    close( mergemod_fd );

  logmsg(1, "mergemem stopped\n");
  free_all();
  exit(0);
}

void logmsg( int lev, char *fmt, ...)
{
  time_t t;
  char dbuf[100];

  if( loglevel >= lev )
  {
    va_list pa;
    va_start( pa, fmt );
    t = time(0);
    (size_t) strftime(dbuf, 100, "%b %d %H:%M:%S", localtime(&t));
    if(logf) fprintf( logf , "%s ", dbuf);
    vfprintf( logf ? logf : stderr, fmt, pa );
    va_end( pa );
    if(logf) fflush(logf);
  }
}

void cfg_reread()
{
  logmsg(1, "rereading cfg-file (SIGHUP)\n");
  signal(SIGHUP, &cfg_reread);
  free_all();
  cmd_list = NULL;
  dont_merge = NULL;
  read_cfg_file();
}

/*
  this is the wait loop which is running in the forked process of the daemon
  it`s not perfect but it does it`s job...
*/
void wait_loop()
{
  struct CmdList *cl;
  unsigned long cntr = 0;

  while(1)
  {
    if(cntr == 0)
    {
      /* merge all commands except dontmerge and commands with own intervals */
      cntr = interval;
      logmsg(2, "merging commands\n");
      merge_commands();
    }
    for(cl = cmd_list; cl; cl=cl->next)
    {
      if(cl->cntr == 0)
      {
	merge_cmd( cl->cmd );
	cl->cntr = cl->iv;
      }
      cl->cntr--;
    }
    cntr--;
    sleep(1);
  }
}

void mrg_singleshot()
{
  struct CmdList *cl;
  /* go through pid list */
  merge_pids( &pid_list );
  for(cl = cmd_list; cl; cl=cl->next)
    merge_cmd( cl->cmd );
}

/*
  merge all commands which are not in the CmdList & DontMerge list...
*/
void merge_commands()
{
  struct CmdList *ctm = NULL, *c, *cl;
  struct DontMerge *dm;
  /*  int found;*/
  /*  struct PidList *pl = NULL, *pl2;*/
  DIR *d;
  struct dirent *de;
  int pid, self;
  FILE *f;
  char path[PATH_MAX+1], comm[1024];
  int rc;

  self = getpid();

  if( !( d = opendir( "/proc" ) ) )
    err_exit( "could not read /proc ?? " );

  while( ( de = readdir( d ) ) )
  {
    if( !( pid = atoi( de->d_name ) ) || pid == self )
      continue;

    sprintf( path, "/proc/%d/stat", pid );
    if( !( f = fopen( path, "r" ) ) )
      continue;

    rc = fscanf( f, "%*d (%[^)]", comm ) == 1;
    fclose( f );

    if( !rc )
      continue;

    for( c = ctm; c != NULL; c = c->next )
    {
      if( strcmp( c->cmd, comm ) == 0)
      {
	c->cntr++;
	break;
      }
    }

    if( c ) /* if found -> next */
      continue;

    /* ensure we dont have this command in the command list*/
    for( cl = cmd_list; cl != NULL; cl = cl->next )
      if( strcmp( cl->cmd, comm ) == 0)
	break;
    if( cl )
      continue;

    /* ensure we dont have this command in the dont merge list*/
    for( dm = dont_merge; dm != NULL; dm = dm->next )
      if( strcmp( dm->cmd, comm ) == 0)
	break;
    if( dm )
      continue;

    if( (c = malloc(sizeof(struct CmdList) + strlen(comm) + 1)) == NULL )
      err_exit("Memory exhausted\n");
    c->cmd = (char*)c+sizeof(struct CmdList);
    strcpy( c->cmd, comm );
    c->next = ctm;
    c->cntr = 1;
    ctm = c;
  }
  closedir( d );

  if( ctm ) /* if more than one pid */
  {
    for( c = ctm; c != NULL; c = c->next )
    {
      if( c->cntr > 1 )
      {
	logmsg( 2, "merging %s (%d instances)\n", c->cmd, c->cntr );
	merge_cmd( c->cmd );
      }
    }
  }

  while(ctm)
  {
    c = ctm;
    ctm = ctm->next;
    free( c );
  }
}

/*
  get all pids of a command and call memory merge routine
*/
void merge_cmd( const char *cmd )
{
  struct PidList *pl = NULL, *pl2;
  DIR *d;
  struct dirent *de;
  int pid, self;
  FILE *f;
  char path[PATH_MAX+1], comm[1024];
  int rc;
  /*  unsigned long found;*/

  logmsg(2, "merging %s\n", cmd);
  self = getpid();

  if( !( d = opendir( "/proc" ) ) )
    err_exit( "could not read /proc ?? " );

  while( ( de = readdir( d ) ) )
  {
    if( !( pid = atoi( de->d_name ) ) || pid == self )
      continue;

    sprintf( path, "/proc/%d/stat", pid );
    if( !( f = fopen( path, "r" ) ) )
      continue;

    rc = fscanf( f, "%*d (%[^)]", comm ) == 1;
    fclose( f );

    if( !rc )
      continue;

    if( strcmp( comm, cmd ) == 0 )
    {
      if( (pl2 = malloc(sizeof(struct PidList)) ) == NULL)
	err_exit( "Memory exhausted\n" );
      pl2->next = pl;
      pl = pl2;
      pl2->pid = pid;
      logmsg( 3, "is running as %d\n", pid );
    }
  }
  closedir( d );

  /* now we have built a list of pids from the command */

  if( pl == NULL ) 
  {
    if( !run_as_daemon )
      logmsg(1, "No processes of command %s found\n", cmd );
    return;
  }

  if( pl->next ) /* if more than one pid */
    merge_pids( &pl );

  while(pl)
  {
    pl2 = pl;
    pl = pl->next;
    free( pl2 );
  }
}

void merge_pids( struct PidList **pl )
{
  if( !*pl ) return;
  if( !(*pl)->next ) return;

  saved_pages = 0;
  chk_runs = 0;
  sum_runs = 0;

  if( init_maps( *pl ) )
  {
    merge_memory( *pl );
  }
  

  logmsg(2, "** Saved Pages: %d is %d KB\n", saved_pages, saved_pages * PAGE_SIZE / 1024);
  logmsg(2, "** check runs: %d\n", chk_runs);
  logmsg(2, "** sum runs: %d\n", sum_runs);

}

int init_maps( struct PidList *pl )
{
  struct PidList *p;
  char buf[100];
  FILE *f;
  unsigned long vm_begin, vm_end;
  int vm_offset, dev_major, dev_minor, inode;
  char line[ MAPLLEN ];
  char vm_perm[ 5 ];
  int i, j;

  if( !pl )
    return FALSE;

  /* read the map files in memory */
  for( p = pl; p != NULL; p = p->next )
  {
    i = j = 0; /* this is the mapcounter */
    logmsg(3, "loading maps for process %d\n", p->pid );
    sprintf( buf, "/proc/%d/maps", p->pid );
    f = fopen( buf, "r" );
    if( !f )
    {
      logmsg(2, "could not open maps file. uuuugh...\n");
      return FALSE;
    }
    while( 1 )
    {
      fgets( line, MAPLLEN, f );
      if( feof( f ) )
	break;
      sscanf( line, "%lx-%lx %s %x %d:%d %d", &vm_begin, &vm_end, vm_perm, &vm_offset,
	      &dev_major, &dev_minor, &inode );
      if( inode == 0 )
      {
	p->vm_begin[i] = vm_begin;
	p->vm_end[i]   = vm_end;
	i++;
      }
      else
	j++;
    }
    logmsg(3, "process has %d anonymous and %d file mappings\n", i, j);
    p->maps = i;
    fclose( f );
  }
  return TRUE;
}

void merge_memory( struct PidList *pl )
{
  int i, k;
  unsigned long *chk;
  unsigned long vm_begin, vm_end;
  int found, found2, vm_begin_smaller, vm_end_bigger;
  struct PidList *p;

  if( !pl ) return;
  if( !pl->next ) return;

  /* reset counter for statistics */
  for( p = pl; p != NULL; p = p->next )
    p->pages_mrg = p->pages_err = p->pages_nmrg = p->pages_ashr = 0;

  /* test if the other pids have the map also and if to correct the beginning and ending */
  for( i = 0; i < pl->maps; i++ )
  {
    vm_begin = pl->vm_begin[i];
    vm_end   = pl->vm_end[i];
    logmsg(3, "comparing mapping %lx-%lx (%ld pages, %ld KB)\n", vm_begin, vm_end,
	   (vm_end-vm_begin) / PAGE_SIZE, (vm_end-vm_begin) / 1024 );

    found2 = 1;
    for( p = pl->next; p != NULL; p=p->next )
    {
      found = 0;
      for( k = 0; k < p->maps; k++ )
      {
	/* correct the beginning if required*/
	if( p->vm_begin[k] <= vm_begin && p->vm_end[k] >= vm_begin )
	{
	  found = 1;
	  vm_begin_smaller = ( p->vm_begin[k] < vm_begin );
	  if( vm_begin_smaller )
	  {
	    vm_begin = p->vm_begin[k];
	    logmsg(3, "decreasing vm_begin to %lx\n", vm_begin );
	  }
	}
	/* correct the ending if required*/
	if( p->vm_end[k] >= vm_end && p->vm_begin[k] <= vm_end )
	{
	  found = 1;
	  vm_end_bigger = ( p->vm_end[k] > vm_end );
	  if( vm_end_bigger )
	  {
	    vm_end = p->vm_end[k];
	    logmsg(3, "increasing vm_end to %lx\n", vm_end );
	  }
	}
      }
      if( !found )
      {
	logmsg(3, "missing mapping ... skipping\n" );
	found2 = 0;
	break;
      }
    }
    if( !found2 )
      continue;

    /* ok, now the map exists in every process and the start/endings have been corrected */
    /* now allocate memory for the checksums and create them for every process */
    for( p = pl; p != NULL; p=p->next )
    {
      if( (chk = malloc( sizeof(unsigned long) * ((vm_end - vm_begin) / PAGE_SIZE + 1))) == NULL)
	err_exit("Memory exhausted\n");                     /* aaaaaarrrggghhh  */
      memset( chk, 0, sizeof(unsigned long) * ((vm_end - vm_begin) / PAGE_SIZE + 1));
      p->chksum = chk;
    }

    /* generate checksums for this mapping */
    gen_checksums( vm_begin, vm_end, pl );

    /* merge memory for this mapping */
    chk_checksums( vm_begin, vm_end, pl );

    for( p = pl; p != NULL; p=p->next )
    {
      free( p->chksum );
      p->pages_mrg += p->mpages_mrg;
      p->pages_err += p->mpages_err;
      p->pages_nmrg += p->mpages_nmrg;
      p->pages_ashr += p->mpages_ashr;
    }
  }

  for( p = pl; p != NULL; p = p->next )
    logmsg(3, "==> pid %d: %dp merged, %dp shared, %dp errors, %dp not merged\n",
	   p->pid, p->pages_mrg, p->pages_ashr, p->pages_err, p->pages_nmrg);
}

/*
  generate an unsigned long checksum for every page & process
*/
void gen_checksums( unsigned long beg, unsigned long end, struct PidList *pl )
{
  struct PidList *p;
  int i = 0;
  unsigned long j;
  struct mergemod_struct mergemod;

  while( beg < end )
  {
    j = 1L;
    for( p = pl; p != NULL; p = p->next )
    {
      
      mergemod.pid1 = p->pid;
      mergemod.addr1 = beg;
      mergemod.pid2 = 0; /* set mode to GET_CHECKSUM */
      if(write(mergemod_fd,&mergemod,sizeof(struct mergemod_struct))
	 == MERGEMEM_SUCCESS)   
	p->chksum[i] = mergemod.addr2;
      else 
	p->chksum[i] = j++;

      /*      printf( "%lx  ", p->chksum[i] );*/
    }
    /*    printf("\n");*/
    i++;
    beg += PAGE_SIZE;
  }
}

/*
  ok this is somewhat a wired function. it checks the checksums of all pages of the
  processes and merges the equal ones. this means if you merge a command with 10 processes
  and the checksum of one page looks like this:

  pid       0  1  2  3  4  5  6  7  8  9
  checksum 10 10 20 30 10 20 20 40 30 30

  then the following pids will be merged:

  1.  (10)  0,1,4
  2.  (20)  2,5,6
  3.  (30)  3,8,9
  4.  (40)  none

  you see, the function tries to be as much efficient as possible :-)

*/
void chk_checksums( unsigned long beg, unsigned long end, struct PidList *pl )
{
  struct PidList *p, *fp;
  int n = 0;
  int fpnew;
  unsigned long *ids;
  int nids;
  int i, j, fnd;
  unsigned long m;
  struct mergemod_struct mergemod;

  for( p = pl; p != NULL; p = p->next )
  {
    p->mpages_mrg = p->mpages_err = p->mpages_nmrg = p->mpages_ashr = 0;
    n++;
  }
  
  if( (ids = (unsigned long *)malloc( sizeof(unsigned long) * n)) == NULL )
    err_exit("Memory exhausted\n");                     /* aaaaaarrrggghhh  */

  chk_runs++;

  i = 0;
  /* walk through the mapping*/
  for( m = beg; m < end; m += PAGE_SIZE )
  {
    nids = 0;
    /* walk through the processes */
    for( p = pl; p != NULL; p = p->next )
    {
      /* look if we found the checksum already */
      fnd = FALSE;
      for( j = 0; j < nids; j++ )
	if( ids[j] == p->chksum[i] )
	{
	  fnd = TRUE;
	  p->id = j;
	  break;
	}
      /* if we didnt find it ... */
      if( !fnd )
      {
	ids[nids] = p->chksum[i];
	p->id = nids;
	nids++;
      }
    }

    /* 
       now all pid structs have set the id var to 0,1,2,3... according to the first second... 
       checksum found. all pids with the same id can be merged 
    */

    for( j = 0; j < nids; j++)
    {
      fp = NULL;
      fpnew = TRUE;
      for( p = pl; p != NULL; p = p->next )
      {
	if( p->id == j )
	{
	  if( fp )
	  {
	    /*
	      BINGO! here we merge...
	    */
	    /*	    printf("%d PID %d + %d  %lx # %d ", i, p->pid, fp->pid,
		   p->chksum[i], p->id);*/
	    mergemod.pid1 = fp->pid;
	    mergemod.addr1 = m;
	    mergemod.pid2 = p->pid;
	    mergemod.addr2 = m;
	    switch( write(mergemod_fd,&mergemod,sizeof(struct mergemod_struct)) )
	      {
	       case MERGEMEM_SUCCESS:
		saved_pages++;
		p->mpages_mrg++;
		if(fpnew) fp->mpages_mrg++;
		/*  if(mergemod.pid1 == p->pid) fp=p;  global order */
		break;
	       case MERGEMEM_ALREADYSH:
		p->mpages_ashr++;
		if(fpnew) fp->mpages_ashr++;
		break;
	       default:
		p->mpages_err++;
		if(fpnew) fp->mpages_err++;
	      }
	    fpnew = FALSE;
	  }
	  else
	    fp = p;
	}
      }
    }

    i++;
  }
  
  for( p = pl; p != NULL; p = p->next )
  {
    p->mpages_nmrg = i - (p->mpages_mrg + p->mpages_ashr + p->mpages_err);
    logmsg(3, "pid %d: %dp merged, %dp shared, %dp errors, %dp not merged\n",
	   p->pid, p->mpages_mrg, p->mpages_ashr, p->mpages_err, p->mpages_nmrg);
  }

  free( ids );
}

void check_running()
{
  DIR *d;
  struct dirent *de;
  int pid, self;
  FILE *f;
  char path[PATH_MAX+1], comm[1024];
  int rc;

  self = getpid();

  if( !( d = opendir( "/proc" ) ) )
    err_exit( "could not read /proc ?? " );

  while( ( de = readdir( d ) ) )
  {
    if( !( pid = atoi( de->d_name ) ) || pid == self )
      continue;

    sprintf( path, "/proc/%d/stat", pid );
    if( !( f = fopen( path, "r" ) ) )
      continue;

    rc = fscanf( f, "%*d (%[^)]", comm ) == 1;
    fclose( f );

    if( !rc )
      continue;

    if( strcmp( comm, "mergemem" ) == 0 )
    {
      printf( "mergemem is already running as pid %d. exiting...\n", pid );
      exit( 1 );
    }
  }
  closedir( d );

}
